# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for JSON schema conversion."""

from typing import Annotated, Literal, Optional
import unittest
from pyglove.core.symbolic import object as pg_object
from pyglove.core.typing import annotation_conversion  # pylint: disable=unused-import
from pyglove.core.typing import json_schema
from pyglove.core.typing import key_specs as ks
from pyglove.core.typing import value_specs as vs


class Foo(pg_object.Object):
  x: int
  y: 'Bar'


class Bar(pg_object.Object):
  z: Optional[Foo]


class JsonSchemaTest(unittest.TestCase):

  maxDiff = None

  def assert_json_schema(
      self, t, expected_json_schema,
      *,
      include_type_name: bool = True,
      include_subclasses: bool = False,
      inline_nested_refs: bool = False):
    actual_schema = json_schema.to_json_schema(
        t,
        include_type_name=include_type_name,
        include_subclasses=include_subclasses,
        inline_nested_refs=inline_nested_refs,
    )
    self.assertEqual(actual_schema, expected_json_schema)

  def test_bool(self):
    self.assert_json_schema(bool, {
        'type': 'boolean',
    })

  def test_int(self):
    self.assert_json_schema(int, {
        'type': 'integer',
    })
    self.assert_json_schema(vs.Int(min_value=0), {
        'type': 'integer',
        'minimum': 0,
    })
    self.assert_json_schema(vs.Int(max_value=1), {
        'type': 'integer',
        'maximum': 1,
    })

  def test_float(self):
    self.assert_json_schema(float, {
        'type': 'number',
    })
    self.assert_json_schema(vs.Float(min_value=0.0), {
        'type': 'number',
        'minimum': 0.0,
    })
    self.assert_json_schema(vs.Float(max_value=1.0), {
        'type': 'number',
        'maximum': 1.0,
    })

  def test_str(self):
    self.assert_json_schema(str, {
        'type': 'string',
    })
    self.assert_json_schema(vs.Str(regex='a.*'), {
        'type': 'string',
        'pattern': 'a.*',
    })

  def test_enum(self):
    self.assert_json_schema(Literal['a', 1], {
        'enum': ['a', 1]
    })
    self.assert_json_schema(Literal['a', 1, None], {
        'anyOf': [
            {'enum': ['a', 1]},
            {'type': 'null'}
        ]
    })
    with self.assertRaisesRegex(
        ValueError, 'Enum candidate .* is not supported'
    ):
      json_schema.to_json_schema(Literal['a', 1, None, ValueError()])

  def test_list(self):
    self.assert_json_schema(vs.List[int], {
        'type': 'array',
        'items': {
            'type': 'integer',
        }
    })

  def test_dict(self):
    self.assert_json_schema(vs.Dict(), {
        'type': 'object', 'additionalProperties': True
    })
    self.assert_json_schema(vs.Dict({'a': int}), {
        'type': 'object',
        'properties': {'a': {
            'type': 'integer',
        }},
        'required': ['a'],
        'additionalProperties': False,
    })
    self.assert_json_schema(vs.Dict([(ks.StrKey(), int)]), {
        'type': 'object',
        'additionalProperties': {'type': 'integer'},
    })
    self.assert_json_schema(vs.Dict([(ks.StrKey(), vs.Any())]), {
        'type': 'object',
        'additionalProperties': True,
    })

  def test_union(self):
    self.assert_json_schema(vs.Union([int, vs.Union([str, int]).noneable()]), {
        'anyOf': [
            {'type': 'integer'},
            {'type': 'string'},
            # TODO(daiyip): Remove duplicates for nested Union in future.
            {'type': 'integer'},
            {'type': 'null'},
        ]
    })

  def test_any(self):
    self.assert_json_schema(vs.Any(), {
        'anyOf': [
            {'type': 'boolean'},
            {'type': 'number'},
            {'type': 'string'},
            {'type': 'array'},
            {'type': 'object', 'additionalProperties': True},
            {'type': 'null'},
        ]
    })

  def test_object(self):
    class A:
      def __init__(self, x: int, y: str):
        pass

    self.assert_json_schema(vs.Object(A), {
        'type': 'object',
        'properties': {
            'x': {
                'type': 'integer',
            },
            'y': {
                'type': 'string',
            },
        },
        'required': ['x', 'y'],
        'title': 'A',
        'additionalProperties': False,
    }, include_type_name=False)

    self.assert_json_schema(vs.Object(A), {
        'type': 'object',
        'properties': {
            'x': {
                'type': 'integer',
            },
            'y': {
                'type': 'string',
            },
        },
        'required': ['x', 'y'],
        'title': 'A',
        'additionalProperties': False,
    }, include_type_name=True)

  def test_pg_object(self):

    class A(pg_object.Object):
      x: int
      y: str

    self.assert_json_schema(vs.Object(A), {
        'type': 'object',
        'properties': {
            'x': {
                'type': 'integer',
            },
            'y': {
                'type': 'string',
            },
        },
        'required': ['x', 'y'],
        'title': 'A',
        'additionalProperties': False,
    }, include_type_name=False)

    self.assert_json_schema(vs.Object(A), {
        'type': 'object',
        'properties': {
            '_type': {
                'const': A.__type_name__,
            },
            'x': {
                'type': 'integer',
            },
            'y': {
                'type': 'string',
            },
        },
        'required': ['_type', 'x', 'y'],
        'title': 'A',
        'additionalProperties': False,
    }, include_type_name=True)

  def test_pg_object_nessted(self):

    class A(pg_object.Object):
      x: Annotated[int, 'field x']
      y: str

    class B(pg_object.Object):
      z: A

    self.assert_json_schema(vs.Object(B), {
        '$defs': {
            'A': {
                'type': 'object',
                'properties': {
                    '_type': {
                        'const': A.__type_name__,
                    },
                    'x': {
                        'type': 'integer',
                        'description': 'field x',
                    },
                    'y': {
                        'type': 'string',
                    },
                },
                'required': ['_type', 'x', 'y'],
                'title': 'A',
                'additionalProperties': False,
            }
        },
        'type': 'object',
        'properties': {
            '_type': {
                'const': B.__type_name__,
            },
            'z': {
                '$ref': '#/$defs/A'
            },
        },
        'required': ['_type', 'z'],
        'title': 'B',
        'additionalProperties': False,
    }, include_type_name=True)

    self.assert_json_schema(vs.Object(B), {
        'type': 'object',
        'properties': {
            '_type': {
                'const': B.__type_name__,
            },
            'z': {
                'type': 'object',
                'properties': {
                    '_type': {
                        'const': A.__type_name__,
                    },
                    'x': {
                        'type': 'integer',
                        'description': 'field x',
                    },
                    'y': {
                        'type': 'string',
                    },
                },
                'required': ['_type', 'x', 'y'],
                'title': 'A',
                'additionalProperties': False,
            },
        },
        'required': ['_type', 'z'],
        'title': 'B',
        'additionalProperties': False,
    }, include_type_name=True, inline_nested_refs=True)

  def test_pg_object_with_subclasses(self):
    class A(pg_object.Object):
      x: int
      y: str

    class B(A):
      z: int

    self.assert_json_schema(
        vs.Object(A).noneable(),
        {
            'anyOf': [
                {
                    'additionalProperties': False,
                    'properties': {
                        '_type': {
                            'const': A.__type_name__,
                        },
                        'x': {
                            'type': 'integer',
                        },
                        'y': {
                            'type': 'string',
                        },
                    },
                    'required': ['_type', 'x', 'y'],
                    'title': 'A',
                    'type': 'object',
                },
                {
                    'additionalProperties': False,
                    'properties': {
                        '_type': {
                            'const': B.__type_name__,
                        },
                        'x': {
                            'type': 'integer',
                        },
                        'y': {
                            'type': 'string',
                        },
                        'z': {
                            'type': 'integer',
                        },
                    },
                    'required': ['_type', 'x', 'y', 'z'],
                    'title': 'B',
                    'type': 'object',
                },
                {
                    'type': 'null',
                }
            ],
        },
        include_type_name=True,
        include_subclasses=True,
    )

  def test_pg_object_with_recursive_refs(self):
    self.assert_json_schema(
        vs.Object(Foo),
        {
            '$defs': {
                'Foo': {
                    'additionalProperties': False,
                    'properties': {
                        '_type': {
                            'const': Foo.__type_name__,
                        },
                        'x': {
                            'type': 'integer',
                        },
                        'y': {
                            '$ref': '#/$defs/Bar',
                        },
                    },
                    'required': ['_type', 'x', 'y'],
                    'title': 'Foo',
                    'type': 'object',
                },
                'Bar': {
                    'additionalProperties': False,
                    'properties': {
                        '_type': {
                            'const': Bar.__type_name__,
                        },
                        'z': {
                            'anyOf': [
                                {'$ref': '#/$defs/Foo'},
                                {'type': 'null'},
                            ]
                        },
                    },
                    'required': ['_type', 'z'],
                    'title': 'Bar',
                    'type': 'object',
                }
            },
            '$ref': '#/$defs/Foo',
        }
    )

  def test_unsupported_value_spec(self):
    with self.assertRaisesRegex(
        TypeError, 'Value spec .* cannot be converted to JSON schema'
    ):
      json_schema.to_json_schema(vs.Callable())

  def test_schema(self):
    class A(pg_object.Object):
      x: int
      y: str

    self.assert_json_schema(
        A.__schema__,
        {
            'type': 'object',
            'properties': {
                '_type': {
                    'const': A.__type_name__,
                },
                'x': {
                    'type': 'integer',
                },
                'y': {
                    'type': 'string',
                },
            },
            'required': ['_type', 'x', 'y'],
            'title': 'A',
            'additionalProperties': False,
        },
        include_type_name=True,
    )

  def test_value_spec_to_json_schema(self):
    self.assertEqual(
        vs.Int().to_json_schema(),
        {
            'type': 'integer',
        }
    )

  def test_schema_to_json_schema(self):
    class A(pg_object.Object):
      x: int
      y: str

    self.assertEqual(
        A.__schema__.to_json_schema(),
        {
            'type': 'object',
            'properties': {
                '_type': {
                    'const': A.__type_name__,
                },
                'x': {
                    'type': 'integer',
                },
                'y': {
                    'type': 'string',
                },
            },
            'required': ['_type', 'x', 'y'],
            'title': 'A',
            'additionalProperties': False,
        }
    )

if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Derived value from other hyper primitives."""

import abc
import copy
from typing import Any, Callable, List, Optional, Tuple, Union

from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core import utils


@symbolic.members([(
    'reference_paths',
    pg_typing.List(pg_typing.Object(utils.KeyPath)),
    (
        'Paths of referenced values, which are relative paths searched from '
        'current node to root.'
    ),
)])
class DerivedValue(symbolic.Object, pg_typing.CustomTyping):
  """Base class of value that references to other values in object tree."""

  @abc.abstractmethod
  def derive(self, *args: Any) -> Any:
    """Derive the value from referenced values."""

  def resolve(
      self, reference_path_or_paths: Optional[Union[str, List[str]]] = None
  ) -> Union[
      Tuple[symbolic.Symbolic, utils.KeyPath],
      List[Tuple[symbolic.Symbolic, utils.KeyPath]],
  ]:
    """Resolve reference paths based on the location of this node.

    Args:
      reference_path_or_paths: (Optional) a string or KeyPath as a reference
        path or a list of strings or KeyPath objects as a list of
        reference paths.
        If this argument is not provided, prebound reference paths of this
        object will be used.

    Returns:
      A tuple (or list of tuple) of (resolved parent, resolved full path)
    """
    single_input = False
    if reference_path_or_paths is None:
      reference_paths = self.reference_paths
    elif isinstance(reference_path_or_paths, str):
      reference_paths = [utils.KeyPath.parse(reference_path_or_paths)]
      single_input = True
    elif isinstance(reference_path_or_paths, utils.KeyPath):
      reference_paths = [reference_path_or_paths]
      single_input = True
    elif isinstance(reference_path_or_paths, list):
      paths = []
      for path in reference_path_or_paths:
        if isinstance(path, str):
          path = utils.KeyPath.parse(path)
        elif not isinstance(path, utils.KeyPath):
          raise ValueError('Argument \'reference_path_or_paths\' must be None, '
                           'a string, KeyPath object, a list of strings, or a '
                           'list of KeyPath objects.')
        paths.append(path)
      reference_paths = paths
    else:
      raise ValueError('Argument \'reference_path_or_paths\' must be None, '
                       'a string, KeyPath object, a list of strings, or a '
                       'list of KeyPath objects.')

    resolved_paths = []
    for reference_path in reference_paths:
      parent = self.sym_parent
      while parent is not None and not reference_path.exists(parent):
        parent = getattr(parent, 'sym_parent', None)
      if parent is None:
        raise ValueError(
            f'Cannot resolve \'{reference_path}\': parent not found.')
      resolved_paths.append((parent, parent.sym_path + reference_path))
    return resolved_paths if not single_input else resolved_paths[0]

  def __call__(self):
    """Generate value by deriving values from reference paths."""
    referenced_values = []
    for reference_path, (parent, _) in zip(
        self.reference_paths, self.resolve()):
      referenced_value = reference_path.query(parent)

      # Make sure referenced value does not have referenced value.
      # NOTE(daiyip): We can support dependencies between derived values
      # in future if needed.
      if not utils.traverse(referenced_value, self._contains_not_derived_value):
        raise ValueError(
            f'Derived value (path={referenced_value.sym_path}) should not '
            f'reference derived values. '
            f'Encountered: {referenced_value}, '
            f'Referenced at path {self.sym_path}.')
      referenced_values.append(referenced_value)
    return self.derive(*referenced_values)

  def _contains_not_derived_value(
      self, path: utils.KeyPath, value: Any
  ) -> bool:
    """Returns whether a value contains derived value."""
    if isinstance(value, DerivedValue):
      return False
    elif isinstance(value, symbolic.Object):
      for k, v in value.sym_items():
        if not utils.traverse(
            v,
            self._contains_not_derived_value,
            root_path=utils.KeyPath(k, path),
        ):
          return False
    return True


class ValueReference(DerivedValue):
  """Class that represents a value referencing another value."""

  def _on_bound(self):
    """Custom init."""
    super()._on_bound()
    if len(self.reference_paths) != 1:
      raise ValueError(
          f'Argument \'reference_paths\' should have exact 1 '
          f'item. Encountered: {self.reference_paths}')

  def derive(self, referenced_value: Any) -> Any:
    """Derive value by return a copy of the referenced value."""
    return copy.copy(referenced_value)

  def custom_apply(
      self,
      path: utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool,
      child_transform: Optional[
          Callable[[utils.KeyPath, pg_typing.Field, Any], Any]
      ] = None,
  ) -> Tuple[bool, 'DerivedValue']:
    """Implement pg_typing.CustomTyping interface."""
    # TODO(daiyip): perform possible static analysis on referenced paths.
    del path, value_spec, allow_partial, child_transform
    return (False, self)


def reference(reference_path: str) -> ValueReference:
  """Create a referenced value from a referenced path."""
  return ValueReference(reference_paths=[reference_path])



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import unittest

from pyglove.core.symbolic import error_info  # pylint: disable=unused-import
from pyglove.core.utils import json_conversion
from pyglove.core.utils import timing


class TimeItTest(unittest.TestCase):
  """Tests for `pg.symbolic.thread_local`."""

  def test_basics(self):
    tc = timing.TimeIt('node')
    self.assertFalse(tc.has_started)
    self.assertEqual(tc.elapse, 0)

    tc.start()
    self.assertTrue(tc.has_started)
    self.assertGreater(tc.elapse, 0)

    self.assertTrue(tc.end())
    self.assertTrue(tc.has_ended)
    elapse1 = tc.elapse
    self.assertFalse(tc.end())
    self.assertEqual(tc.elapse, elapse1)

  def test_timeit(self):
    with timing.timeit('node') as t:
      self.assertEqual(t.name, 'node')
      self.assertIsNotNone(t.start_time)
      self.assertTrue(t.has_started)
      self.assertIsNone(t.end_time)
      self.assertFalse(t.has_ended)
      time.sleep(0.5)
      elapse1 = t.elapse
      self.assertEqual(t.children, [])
      with timing.timeit('child') as t1:
        time.sleep(0.5)
        with timing.timeit('grandchild') as t2:
          time.sleep(0.5)

        self.assertEqual(t.children, [t1])
        self.assertEqual(t1.children, [t2])

        r = t.status()
        self.assertTrue(r['node'].has_started)
        self.assertGreater(r['node'].elapse, 0)
        self.assertFalse(r['node'].has_ended)
        self.assertFalse(r['node'].has_error)
        self.assertFalse(r['node.child'].has_ended)
        self.assertTrue(r['node.child.grandchild'].has_ended)

        with timing.timeit('grandchild') as t3:
          time.sleep(0.5)
        self.assertEqual(t1.children, [t2, t3])

    elapse2 = t.elapse
    self.assertTrue(t.has_ended)
    self.assertGreater(elapse2, elapse1)
    time.sleep(0.5)
    self.assertEqual(elapse2, t.elapse)

    status = t.status()
    self.assertEqual(
        list(status.keys()),
        ['node', 'node.child', 'node.child.grandchild']
    )
    self.assertEqual(
        status['node.child.grandchild'].elapse, t2.elapse + t3.elapse
    )
    self.assertEqual(
        sorted([v.elapse for v in status.values()], reverse=True),
        [v.elapse for v in status.values()],
    )
    self.assertTrue(all(v.has_ended for v in status.values()))
    self.assertFalse(any(v.has_error for v in status.values()))
    status = t.status()
    json_dict = json_conversion.to_json(status)
    status2 = json_conversion.from_json(json_dict)
    self.assertIsNot(status2, status)
    self.assertEqual(status2, status)

  def test_timeit_with_error(self):
    with self.assertRaises(ValueError):
      with timing.timeit('node') as t:
        with timing.timeit('child') as t1:
          with timing.timeit('grandchild') as t2:
            raise ValueError('error')

    r = t.status()
    self.assertTrue(r['node'].has_error)
    self.assertTrue(t.has_error)
    self.assertTrue(t.error.tag.startswith('ValueError'))
    self.assertTrue(r['node'].error.tag.startswith('ValueError'))
    self.assertTrue(r['node.child'].has_error)
    self.assertTrue(t1.has_error)
    self.assertTrue(r['node.child.grandchild'].has_error)
    self.assertTrue(t2.has_error)

  def test_timeit_summary(self):
    summary = timing.TimeIt.StatusSummary()
    self.assertFalse(summary)
    for i in range(10):
      with timing.timeit() as t:
        time.sleep(0.1)
        with timing.timeit('child'):
          time.sleep(0.1)
          try:
            with timing.timeit('grandchild'):
              time.sleep(0.1)
              if i < 2:
                raise ValueError('error')
          except ValueError:
            pass
      summary.aggregate(t.status())
    self.assertTrue(summary)
    self.assertEqual(
        list(summary.breakdown.keys()),
        ['', 'child', 'child.grandchild']
    )
    self.assertEqual(
        [x.num_started for x in summary.breakdown.values()],
        [10, 10, 10]
    )
    self.assertEqual(
        [x.num_ended for x in summary.breakdown.values()],
        [10, 10, 10]
    )
    self.assertEqual(
        [x.num_failed for x in summary.breakdown.values()],
        [0, 0, 2]
    )
    self.assertEqual(
        summary.breakdown['child.grandchild'].error_tags,
        {'ValueError': 2},
    )
    # Test serialization.
    json_dict = summary.to_json()
    summary2 = timing.TimeIt.StatusSummary.from_json(json_dict)
    self.assertIsNot(summary2, summary)
    self.assertEqual(summary2.breakdown, summary.breakdown)


if __name__ == '__main__':
  unittest.main()
